import numpy

from dexp.io.compress_array import compress_array, decompress_array


def do_test(array_dc, array_uc, min_num_chunks=0):
    compressed_array = compress_array(array_uc,
                                      min_num_chunks=min_num_chunks)
    ratio = len(compressed_array) / (array_uc.size * array_uc.itemsize)
    print("Compression succeeded!")
    print(f"Compression ratio: {ratio}")
    decompress_array(compressed_array, array_dc)
    print("Decompression succeeded!")
    assert (array_uc == array_dc).all()

def test_compress_huge_array():
    array_uc = numpy.linspace(0, 1, 2**33).astype(numpy.uint16)
    array_dc = numpy.empty_like(array_uc)
    do_test(array_dc, array_uc)

def test_compress_large_array():
    array_uc = numpy.linspace(0, 1024, 2**30).astype(numpy.uint16)
    array_dc = numpy.empty_like(array_uc)
    do_test(array_dc, array_uc)

def test_compress_small_array():
    array_uc = numpy.linspace(0, 1024, 2**15).astype(numpy.uint16)
    array_dc = numpy.empty_like(array_uc)
    do_test(array_dc, array_uc)

def test_compress_tiny_array():
    array_uc = numpy.linspace(0, 1024, 1)
    array_dc = numpy.empty_like(array_uc).astype(numpy.uint16)
    do_test(array_dc, array_uc)

def test_compress_many_chunks():
    array_uc = numpy.linspace(0, 1024, 1000).astype(numpy.uint16)
    array_dc = numpy.empty_like(array_uc)
    do_test(array_dc, array_uc, min_num_chunks=987)





import cupy
import numpy as np
from typing import List

from arbol.arbol import aprint
from arbol.arbol import asection
from zarr.errors import ContainsArrayError, ContainsGroupError

import dask
from dask.distributed import Client
from dask_cuda import LocalCUDACluster

from dexp.processing.backends.backend import Backend
from dexp.processing.backends.best_backend import BestBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.multiview_lightsheet.fusion.mvsols import msols_fuse_1C2L
from dexp.processing.multiview_lightsheet.fusion.simview import SimViewFusion
from dexp.processing.registration.model.model_io import model_list_from_file, model_list_to_file


def dataset_fuse(dataset,
                 output_path,
                 channels,
                 slicing,
                 store,
                 compression,
                 compression_level,
                 overwrite,
                 microscope,
                 equalise,
                 equalise_mode,
                 zero_level,
                 clip_too_high,
                 fusion,
                 fusion_bias_strength_i,
                 fusion_bias_strength_d,
                 dehaze_size,
                 dark_denoise_threshold,
                 z_pad_apodise,
                 loadreg,
                 model_list_filename,
                 warpreg_num_iterations,
                 min_confidence,
                 max_change,
                 registration_edge_filter,
                 maxproj,
                 huge_dataset,
                 workers,
                 workersbackend,
                 devices,
                 check,
                 stop_at_exception=True):

    views = tuple(dataset.get_array(channel, per_z_slice=False) for channel in channels)

    with asection(f"Views:"):
        for view, channel in zip(views, channels):
            aprint(f"View: {channel} of shape: {view.shape} and dtype: {view.dtype}")

    output_shape = views[0][slicing].shape
    total_time_points = views[0].shape[0]
    time_points = list(range(total_time_points))
    if slicing is not None:
        aprint(f"Slicing with: {slicing}")
        if isinstance(slicing, tuple):
            time_points = time_points[slicing[0]]
            slicing = slicing[1:]
        else:  # slicing only over time
            time_points = time_points[slicing]
            slicing = ...
    else:
        slicing = ...

    dtype = views[0].dtype

    # We allocate last minute once we know the shape...
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(output_path, mode, store)

    # load registration models:
    with NumpyBackend():
        if loadreg:
            aprint(f"Loading registration shifts from existing file! ({model_list_filename})")
            models = model_list_from_file(model_list_filename)
        else:
            models = [None] * len(time_points)

    print('creating dataset', output_shape)
    if 'fused' not in dest_dataset.channels():
        try:
            dest_dataset.add_channel('fused',
                                     shape=output_shape,
                                     dtype=dtype,
                                     codec=compression,
                                     clevel=compression_level)
        except (ContainsArrayError, ContainsGroupError):
                aprint(f"Other thread/process created channel before... ")
    print('done')

    @dask.delayed
    def process(i, params):
        equalisation_ratios_reference, model = params
        tp = time_points[i]
        try:
            with asection(f"Loading channels {channels} for time point {i}/{len(time_points)}"):
                views_tp = tuple(np.asarray(view[tp][slicing]) for view in views)

            with BestBackend(exclusive=True, enable_unified_memory=True):
                if models[i] is not None:
                    model = models[i]
                # otherwise it could be a None model or from the first iteration if equalisation mode was 'first'

                if microscope == 'simview':
                    fuse_obj = SimViewFusion(registration_model=model,
                                             equalise=equalise,
                                             equalisation_ratios=equalisation_ratios_reference,
                                             zero_level=zero_level,
                                             clip_too_high=clip_too_high,
                                             fusion=fusion,
                                             fusion_bias_exponent=2,
                                             fusion_bias_strength_i=fusion_bias_strength_i,
                                             fusion_bias_strength_d=fusion_bias_strength_d,
                                             dehaze_before_fusion=True,
                                             dehaze_size=dehaze_size,
                                             dehaze_correct_max_level=True,
                                             dark_denoise_threshold=dark_denoise_threshold,
                                             dark_denoise_size=9,
                                             butterworth_filter_cutoff=1,
                                             flip_camera1=True)

                    tp_array = fuse_obj(*views_tp)
                    new_equalisation_ratios = fuse_obj._equalisation_ratios

                elif microscope == 'mvsols':
                    metadata = dataset.get_metadata()
                    angle = metadata['angle']
                    channel = metadata['channel']
                    dz = metadata['dz']
                    res = metadata['res']
                    illumination_correction_sigma = metadata['ic_sigma'] if 'ic_sigma' in metadata else None

                    tp_array, model, new_equalisation_ratios = msols_fuse_1C2L(*views_tp,
                                                                               z_pad=z_pad_apodise[0],
                                                                               z_apodise=z_pad_apodise[1],
                                                                               registration_num_iterations=warpreg_num_iterations,
                                                                               registration_force_model=loadreg,
                                                                               registration_model=model,
                                                                               registration_min_confidence=min_confidence,
                                                                               registration_max_change=max_change,
                                                                               registration_edge_filter=registration_edge_filter,
                                                                               equalise=equalise,
                                                                               equalisation_ratios=equalisation_ratios_reference[0],
                                                                               zero_level=zero_level,
                                                                               clip_too_high=clip_too_high,
                                                                               fusion=fusion,
                                                                               fusion_bias_exponent=2,
                                                                               fusion_bias_strength_x=fusion_bias_strength_i,
                                                                               dehaze_size=dehaze_size,
                                                                               dark_denoise_threshold=dark_denoise_threshold,
                                                                               angle=angle,
                                                                               dx=res,
                                                                               dz=dz,
                                                                               illumination_correction_sigma=illumination_correction_sigma,
                                                                               registration_mode='projection' if maxproj else 'full',
                                                                               huge_dataset_mode=huge_dataset)
                else:
                    raise NotImplementedError

                with asection(f"Moving array from backend to numpy."):
                    tp_array = Backend.to_numpy(tp_array, dtype=dtype, force_copy=False)

                model = model.to_numpy()
                del views_tp
                Backend.current().clear_memory_pool()

                aprint(f"Last equalisation ratios: {new_equalisation_ratios}")

            with asection(f"Saving fused stack for time point {i}, shape:{tp_array.shape}, dtype:{tp_array.dtype}"):
                dest_dataset.write_stack(channel='fused',
                                         time_point=i,
                                         stack_array=tp_array)

            aprint(f"Done processing time point: {i}/{len(time_points)} .")

        except Exception as error:
            aprint(error)
            aprint(f"Error occurred while processing time point {i} !")
            import traceback
            traceback.print_exc()

            if stop_at_exception:
                raise error

        return new_equalisation_ratios, model

    cluster = LocalCUDACluster(CUDA_VISIBLE_DEVICES=devices)
    client = Client(cluster)
    aprint('Dask Client', client)

    if microscope == 'simview':
        params = ([None, None, None], None)
    elif microscope == 'mvsols':
        params = ([None], None)
    else:
        raise NotImplementedError

    # if equalise mode equals 'first' it creates a dependency to the first computation
    lazy_computations = []
    if equalise_mode == 'first':
        params = process(0, params)
        lazy_computations.append(params)

    for i in range(1 if equalise_mode == 'first' else 0, len(time_points)):
        lazy_computations.append(process(i, params))

    models = [model for _, model in dask.compute(*lazy_computations)]

    if not loadreg:
        model_list_to_file(model_list_filename, models)

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    dest_dataset.set_cli_history(parent=dataset if isinstance(dataset, ZDataset) else None)
    # close destination dataset:
    dest_dataset.close()



from typing import Union
from dexp.processing.registration.model.pairwise_registration_model import PairwiseRegistrationModel
from dexp.processing.registration.model.sequence_registration_model import SequenceRegistrationModel

RegistrationModel = Union[SequenceRegistrationModel, PairwiseRegistrationModel]


import os
from typing import Sequence, Optional

import numpy
from arbol.arbol import aprint, asection

from dexp.datasets.base_dataset import BaseDataset


def dataset_copy(dataset: BaseDataset,
                 dest_path: str,
                 channels: Sequence[str],
                 slicing,
                 store: str,
                 chunks: Optional[Sequence[int]],
                 compression: str,
                 compression_level: int,
                 overwrite: bool,
                 zerolevel: int,
                 workers: int,
                 workersbackend: str,
                 check: bool,
                 stop_at_exception: bool = True,
                 ):

    # Create destination dataset:
    from dexp.datasets.zarr_dataset import ZDataset
    mode = 'w' + ('' if overwrite else '-')
    dest_dataset = ZDataset(dest_path, mode, store)

    # Process each channel:
    for channel in dataset._selected_channels(channels):
        with asection(f"Copying channel {channel}:"):
            array = dataset.get_array(channel)

            total_time_points = shape[0]
            time_points = list(range(total_time_points))
            if slicing is not None:
                aprint(f"Slicing with: {slicing}")
                if isinstance(slicing, tuple):
                    time_points = time_points[slicing[0]]
                    slicing = slicing[1:]
                else:  # slicing only over time
                    time_points = time_points[slicing]
                    slicing = ...
            else:
                slicing = ...

            shape = array[0][slicing].shape
            dtype = array.dtype
            if chunks is None:
                chunks = ZDataset._default_chunks

            dest_dataset.add_channel(name=channel,
                                     shape=shape,
                                     dtype=dtype,
                                     chunks=chunks,
                                     codec=compression,
                                     clevel=compression_level)

            def process(i):
                tp = time_points[i]
                try:
                    aprint(f"Processing time point: {i} ...")
                    tp_array = array[tp][slicing]
                    if zerolevel != 0:
                        tp_array = numpy.array(tp_array)
                        tp_array = numpy.clip(tp_array, a_min=zerolevel, a_max=None, out=tp_array)
                        tp_array -= zerolevel
                    dest_dataset.write_stack(channel=channel,
                                             time_point=i,
                                             stack_array=tp_array)
                except Exception as error:
                    aprint(error)
                    aprint(f"Error occurred while copying time point {i} !")
                    import traceback
                    traceback.print_exc()
                    if stop_at_exception:
                        raise error

            for i in range(len(time_points)):
                process(i)

    aprint(dest_dataset.info())
    if check:
        dest_dataset.check_integrity()

    dest_dataset.set_cli_history(parent=dataset if isinstance(dataset, ZDataset) else None)
    # close destination dataset:
    dest_dataset.close()



import os

import click
from arbol.arbol import aprint, asection

from dexp.cli.parsing import _get_output_path
from dexp.utils.robocopy import robocopy


@click.command()
@click.argument('input_path', nargs=1)  # ,  help='input path'
@click.option('--output_path', '-o')  # , help='output path'
@click.option('--workers', '-wk', default=8, help='Number of worker threads to spawn. Negative numbers n correspond to: number_of _cores / |n| ', show_default=True)  #
@click.option('--large_files', '-lf', is_flag=True, help='Set to true to speed up large file transfer', show_default=True)
def fastcopy(input_path,
             output_path,
             workers,
             large_files):
    """ Copies a dataset fast, with no processing, just moves the data as fast as possible. For each operating system it uses the best method.
    """

    output_path = _get_output_path(input_path, output_path, '_copy')

    if workers == -1:
        workers = max(1, os.cpu_count() // abs(workers))

    with asection(f"Fast copying from: {input_path} to {output_path} "):

        from sys import platform
        if platform == "linux" or platform == "linux2":
            raise NotImplementedError("Fast copy not yet implemented for Linux")
        elif platform == "darwin":
            raise NotImplementedError("Fast copy not yet implemented for OSX")
        elif platform == "win32":
            robocopy(input_path, output_path, nb_threads=workers, large_files=large_files or not('.zarr' in input_path))

        aprint("Done!")




import multiprocessing
from multiprocessing.pool import ThreadPool

import dask
import numcodecs
from arbol import aprint
from napari.utils.dask_utils import create_dask_cache
from numcodecs import blosc

def config_dask_and_blosc():
    # Configure multithreading for Dask:
    _cpu_count = multiprocessing.cpu_count() // 2
    # aprint(f"Number of cores on system: {_cpu_count}")
    _nb_threads = max(1, _cpu_count)

    aprint(f"Number of threads used by BLOSC and DASK: {blosc.get_nthreads()}")

    dask.config.set(scheduler='threads')
    dask.config.set(pool=ThreadPool(_nb_threads))

    # Configure multithreading for Blosc:
    blosc.set_nthreads(_nb_threads)

    # same for numcodecs:
    numcodecs.blosc.use_threads = True
    numcodecs.blosc.set_nthreads(_nb_threads)

    #Config Dask cache using napari:
    try:
        create_dask_cache(mem_fraction=0.95)
    except Exception as e:
        aprint("Warning: Setting Dask caching failed.")


import numpy

from dexp.optics.psf.standard_psfs import nikon16x08na
from dexp.processing.backends.backend import Backend
from dexp.processing.backends.cupy_backend import CupyBackend
from dexp.processing.backends.numpy_backend import NumpyBackend
from dexp.processing.deconvolution.lr_deconvolution import lucy_richardson_deconvolution
from dexp.processing.filters.fft_convolve import fft_convolve
from dexp.processing.synthetic_datasets.nuclei_background_data import generate_nuclei_background_data
from dexp.utils.timeit import timeit


def demo_lr_deconvolution_numpy():
    with NumpyBackend():
        _demo_lr_deconvolution()


def demo_lr_deconvolution_cupy():
    try:
        with CupyBackend():
            _demo_lr_deconvolution()
    except ModuleNotFoundError:
        print("Cupy module not found! Test passes nevertheless!")


def _demo_lr_deconvolution(length_xy=128):
    xp = Backend.get_xp_module()

    with timeit("generate data"):
        image_gt, background, image = generate_nuclei_background_data(add_noise=False,
                                                                      length_xy=length_xy,
                                                                      length_z_factor=1,
                                                                      zoom=2,
                                                                      background_stength=0,
                                                                      add_offset=False)
    psf_size = 31

    psf = nikon16x08na(xy_size=psf_size,
                       z_size=psf_size)
    # psf = olympus20x10na()
    psf = psf.astype(dtype=numpy.float16)
    image = image.astype(dtype=numpy.float16)

    blurry = fft_convolve(image, psf)
    blurry = blurry - blurry.min()
    blurry = blurry / blurry.max()
    noisy = blurry
    noisy = noisy + 0.01 * xp.random.uniform(size=blurry.shape)

    max_correction = 8

    iterations = 50

    iterations_wb = 5
    wb_cutoffs = 0.9
    wb_order = 2
    wb_beta = 0.05

    with timeit("deconvolved_wb"):
        deconvolved_wb = lucy_richardson_deconvolution(noisy, psf,
                                                       num_iterations=iterations_wb,
                                                       back_projection='wb',
                                                       wb_cutoffs=wb_cutoffs,
                                                       wb_beta=wb_beta,
                                                       wb_order=wb_order,
                                                       max_correction=max_correction,
                                                       padding=psf_size)

    with timeit("deconvolved_same"):
        deconvolved_same = lucy_richardson_deconvolution(noisy, psf,
                                                         num_iterations=iterations_wb,
                                                         max_correction=max_correction,
                                                         padding=psf_size)

    with timeit("deconvolved"):
        deconvolved = lucy_richardson_deconvolution(noisy, psf,
                                                    num_iterations=iterations,
                                                    max_correction=max_correction,
                                                    padding=psf_size)

    # with timeit("lucy_richardson_deconvolution (scatter-gather)"):
    #     def f(_image):
    #         return lucy_richardson_deconvolution(backend, _image, psf,
    #                                              num_iterations=iterations_wb,
    #                                              back_projection='wb',
    #                                              wb_beta=wb_beta,
    #                                              wb_order=wb_order,
    #                                              max_correction=max_correction,
    #                                              normalise_input=False,
    #                                              padding=16)
    #
    #     deconvolved_sg_wb = scatter_gather_i2i(backend, f, noisy,
    #                                         chunks=length_xy // 2,
    #                                         margins=17,
    #                                         normalise=True,
    #                                         to_numpy=True)

    from napari import Viewer, gui_qt
    with gui_qt():
        def _c(array):
            return Backend.to_numpy(array)

        viewer = Viewer()
        viewer.add_image(_c(image), name='image')
        viewer.add_image(_c(blurry), name='blurry')
        viewer.add_image(_c(psf), name='psf')
        viewer.add_image(_c(noisy), name='noisy')
        viewer.add_image(_c(deconvolved), name='deconvolved')
        viewer.add_image(_c(deconvolved_same), name='deconvolved_same')
        viewer.add_image(_c(deconvolved_wb), name='deconvolved_wb')
        # viewer.add_image(_c(deconvolved_sg_wb), name='deconvolved_sg_wb')


if __name__ == "__main__":
    demo_lr_deconvolution_cupy()
    demo_lr_deconvolution_numpy()



from math import ceil

import blosc
from blosc import get_cbuffer_sizes, decompress, decompress_ptr, compress_ptr
from numpy import ndarray


def compress_array(array: ndarray, clevel: int = 3, compressor: str = 'lz4', min_num_chunks: int = 0) -> bytes:
    """
    Compresses an arbitrary ndarray that supports the '__array_interface__' into a Blosc compressed buffer.

    Parameters
    ----------
    array: Array to compress
    clevel: Compression level
    compressor: compressor (any supported by Blosc)
    min_num_chunks: Minimum number of chunks to split array into before compression.

    Returns
    -------
    Compressed buffer (as bytes)

    """
    array_address = array.__array_interface__['data'][0]
    array_length = array.nbytes
    end_address = array_address+array_length
    itemsize = array.itemsize

    # Blosc does not support 64bit buffers, only 32 bit buffers:
    max_chunk_length = blosc.MAX_BUFFERSIZE

    # Let's avoid small trailing chunks and make them all about teh same size:
    number_of_chunks = max(2, ceil(array_length/max_chunk_length)) if array_length > max_chunk_length else 1
    number_of_chunks = max(min_num_chunks, number_of_chunks)
    approx_chunk_length_in_bytes = array_length//number_of_chunks

    # Let's make sure that the chunks are aligned to the data type:
    approx_chunk_length_in_bytes = (approx_chunk_length_in_bytes//itemsize)*itemsize

    # Let' make the chunks large enough that we are garanteed to cover the whole array:
    approx_chunk_length_in_bytes += number_of_chunks*itemsize

    # this will hold the compressed data:
    compressed_buffer = bytearray()

    # we start reading the array data here:
    chunk_address = array_address

    # We loop through each chunk:
    while chunk_address < end_address:
        chunk_length_in_bytes = min(approx_chunk_length_in_bytes, end_address-chunk_address)
        compressed_chunk = compress_ptr(address=chunk_address,
                     items=chunk_length_in_bytes//itemsize,
                     typesize=itemsize,
                     clevel=clevel,
                     shuffle=blosc.BITSHUFFLE,
                     cname=compressor)
        if number_of_chunks == 1:
            # If there is only one chunk, let's not be complicated about it:
            compressed_buffer = compressed_chunk
        else:
            # If not, so bit it:
            compressed_buffer.extend(compressed_chunk)
        chunk_address += chunk_length_in_bytes

    return bytes(compressed_buffer)

def decompress_array(compressed_bytes, out_array: ndarray = None) -> ndarray:
    """
    Decompresses an array compressed with the 'compress_array' function.

    Parameters
    ----------
    compressed_bytes: buffer containing compressed data
    out_array: array of _correct_ size to put the decompressed daat in.

    Returns
    -------
    Same array as passed as 'out_array'

    """

    # get the array pointer and length in bytes:
    array_address = out_array.__array_interface__['data'][0]
    array_length = out_array.nbytes

    # prepare the basics:
    num_of_compressed_bytes = len(compressed_bytes)
    offset_compressed = 0
    address_decompressed = array_address

    while (num_of_compressed_bytes - offset_compressed > 32):

        # This is the BLOSC header:
        blosc_header = bytes(compressed_bytes[offset_compressed:offset_compressed + 32])

        # we check how large is the first chunk available:
        num_decompressed_bytes, num_compressed_bytes, blocksize = get_cbuffer_sizes(blosc_header)

        # we prepare the corresponding region of the array:
        compressed_chunk = compressed_bytes[offset_compressed:offset_compressed+num_compressed_bytes]

        # do the actual decompression:
        decompress_ptr(compressed_chunk, address_decompressed)

        # we move the pointers forward:
        offset_compressed += num_compressed_bytes
        address_decompressed += num_decompressed_bytes

    return out_array







import os
import subprocess

from arbol import aprint, section, asection


def robocopy(source_folder: str, dest_folder: str, move_files: bool = False, nb_threads: int = 8, large_files: bool = False, wait_to_finish: bool = True):
    '''
        Start a generic robocopy job on Windows to copy files from one folder to another

        Args
        ---------

        Returns
        ---------
        copyProcess (Popen Object) :
        > this is the Popen object that's running the copy process. This can be
        > used to check or kill the process if necessary.
    '''

    with asection(f"Starting a Windows Robocopy job: copying all files and folders from {source_folder} to {dest_folder} with {nb_threads} threads."):
        # checks for network file paths in source and destination paths
        if "\\" in source_folder[:1]:
            source_folder = "\\" + source_folder

        if "\\" in dest_folder[:1]:
            dest_folder = "\\" + dest_folder

        # replaces slashes:
        source                      = (source_folder).replace('/', '\\')
        dest                        = (dest_folder).replace('/', '\\')

        # logfile:
        log_file = 'robocopy_log.txt'
        if os.path.exists(log_file):
            os.remove(log_file)

        # format the command list for Popen
        copyCommand                 = ["ROBOCOPY",
                                        source,
                                        dest,
                                       '/e',
                                       '/R:10',
                                       '/W:5',
                                       '/TBD',
                                       '/NP',
                                       #'/V',
                                       '/eta',
                                       '/tee',
                                       f'/MT:{nb_threads}',
                                       f'/log:{log_file}'
                                       ]

        if move_files:
            copyCommand.append('/move')

        if large_files:
            copyCommand.append('/j')

         # use subprocess to start a copy command
        aprint(f"Robocopy command: {copyCommand}")
        copy_process = subprocess.Popen(copyCommand)

        if wait_to_finish:
            copy_process.wait()

    return copy_process